/*

sudosh - sudo shell that supports input and output logging to syslog

Copyright 2004 and $Date: 2004/10/18 01:37:11 $
                Douglas Richard Hanks Jr.

Licensed under the Open Software License version 2.0

This program is free software; you can redistribute it
and/or modify it under the terms of the Open Software
License, version 2.0 by Lauwrence E. Rosen.

This program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the Open Software License for details.

*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <paths.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/signal.h>
#include <errno.h>
#include <syslog.h>
#include <pwd.h>
#include <sys/wait.h>

#include <unistd.h>
#include <string.h>

#include "config.h"

void finish(int);
void done(void);
void fail(void);
void resize(int);
void fixtty(void);
void getmaster(void);
void getslave(void);
void doinput(void);
void dooutput(void);
void doshell(void);
void mysyslog(int, const char *, ...);

char *shell;
FILE *fscript;
FILE *ftime;
FILE *finput;
int master;
int slave;
int child;
int subchild;

struct termios tt;
struct winsize win;
int lb;
int l;
char line[] = "/dev/ptyXX";

static char *progname;

int main(int argc, char **argv)
{
    extern int optind;
    int ch;
    char *p = NULL;
    char *user = getlogin();
    char *sudo_user = NULL;
    char *vshell = NULL;
    char lscript[BUFSIZ];
    char ltime[BUFSIZ];
    char linput[BUFSIZ];
    struct stat d, s;
    time_t now;
    int valid = -1;

    progname = argv[0];
    if ((p = strrchr(progname, '/')) != NULL)
	progname = p + 1;


    if (argc == 2) {
	if (!strcmp(argv[1], "-V") || !strcmp(argv[1], "--version")) {
	    printf("%s from %s\n", progname, VERSION);
	    return 0;
	}
    }


    if ((sudo_user = getenv("SUDO_USER")) == NULL) {
	fprintf(stderr,
		"sudosh must be called via sudo.  i.e. \"sudo sudosh\"\n");
	exit(255);
    }

    if (user == NULL) {
	fprintf(stderr,
		"getlogin() returned NULL.  I do not know who you are.  Stopping.\n");
	exit(255);
    }

    if (stat((const char *) LOGDIR, &d) == -1) {
	if (mkdir((const char *) LOGDIR, S_IRWXU) == -1) {
	    char str[BUFSIZ];

	    snprintf(str, BUFSIZ - 1, "mkdir(%s)", LOGDIR);
	    perror(str);
	    fprintf(stderr,
		    "Directory %s needs to exist and be created by root with the permissions of 0700\n");
	    exit(errno);
	}
    }

    fprintf(stderr, "mode=%i\n", d.st_mode);
    if ((shell = getenv("SHELL")) == NULL)
	shell = getpwnam(user)->pw_shell;


    /* check against /etc/shells to make sure it's a real shell */
    setusershell();
    while (vshell = getusershell()) {
	if (strcmp(shell, vshell) == 0)
	    valid = 1;
    }
    endusershell();

    if (valid != 1) {
	if (shell == NULL) {
	    fprintf(stderr, "Could not determine a valid shell.\n");
	    mysyslog(LOG_ERR, "Could not determine a valid shell");
	    exit(200);
	} else {
	    fprintf(stderr, "%s is not in /etc/shells\n", shell);
	    mysyslog(LOG_ERR, "%s,%s: %s is not in /etc/shells",
		     user, ttyname(0), shell);
	    exit(210);
	}
    }

    if (stat((const char *) shell, &s) == -1) {
	fprintf(stderr, "Shell %s doesn't exist.\n", shell);
	mysyslog(LOG_ERR, "%s,%s: shell %s doesn't exist.",
		 user, ttyname(0), shell);
	exit(255);
    }

    now = time((time_t *) NULL);

    snprintf(lscript, (size_t) BUFSIZ - 1,
	     "%s/%s-script-%i", LOGDIR, user, now);
    snprintf(ltime, (size_t) BUFSIZ - 1,
	     "%s/%s-time-%i", LOGDIR, user, now);
    snprintf(linput, (size_t) BUFSIZ - 1,
	     "%s/%s-input-%i", LOGDIR, user, now);

    if ((fscript = fopen(lscript, "w")) == NULL) {
	fprintf(stderr, "fopen(): %s\n", lscript);
	perror("fopen");
	exit(200);
    }
    if ((ftime = fopen(ltime, "w")) == NULL) {
	fprintf(stderr, "fopen(): %s\n", ltime);
	perror("fopen");
	exit(200);
    }
    if ((finput = fopen(linput, "w")) == NULL) {
	fprintf(stderr, "fopen(): %s\n", linput);
	perror("fopen");
	exit(200);
    }


    getmaster();
    fixtty();

    (void) signal(SIGCHLD, finish);
    child = fork();
    if (child < 0) {
	perror("fork");
	fail();
    }
    if (child == 0) {
	subchild = child = fork();
	if (child < 0) {
	    perror("fork");
	    fail();
	}
	if (child)
	    dooutput();
	else
	    doshell();
    } else
	(void) signal(SIGWINCH, resize);
    doinput();

    return 0;
}

void doinput()
{
    register int cc;
    char ibuf[BUFSIZ];

    (void) fclose(fscript);

    while ((cc = read(0, ibuf, BUFSIZ)) > 0) {
	(void) write(master, ibuf, cc);

	switch (*ibuf) {
	case '\r':
	    fprintf(finput, "\n");
	    break;
	case 0x003:
	    fprintf(finput, "(CTRL-C)");
	    break;
	case 0x004:
	    fprintf(finput, "(CTRL-D)\n");
	    break;
	case 0x1a:
	    fprintf(finput, "(CTRL-Z)\n");
	    break;

	case 0x1b:
	    fprintf(finput, "(ESC)");
	    break;
	default:
	    (void) fwrite(ibuf, 1, cc, finput);
	    break;
	}
    }
    done();
}


void finish(int dummy)
{
    int status;
    register int pid;
    register int die = 0;

    while ((pid = wait3(&status, WNOHANG, 0)) > 0)
	if (pid == child)
	    die = 1;

    if (die)
	done();
}

void resize(int dummy)
{
    /* transmit window change information to the child */
    (void) ioctl(0, TIOCGWINSZ, (char *) &win);
    (void) ioctl(slave, TIOCSWINSZ, (char *) &win);

    kill(child, SIGWINCH);
}

/*
 * Stop extremely silly gcc complaint on %c:
 *  warning: `%c' yields only last 2 digits of year in some locales
 */
static void
my_strftime(char *buf, size_t len, const char *fmt, const struct tm *tm)
{
    strftime(buf, len, fmt, tm);
}

void dooutput()
{
    register int cc;
    char obuf[BUFSIZ];
    struct timeval tv;
    double oldtime = time(NULL), newtime;

    (void) close(0);

    for (;;) {
	gettimeofday(&tv, NULL);
	cc = read(master, obuf, sizeof(obuf));
	if (cc <= 0)
	    break;
	newtime = tv.tv_sec + (double) tv.tv_usec / 1000000;
	fprintf(ftime, "%f %i\n", newtime - oldtime, cc);
	oldtime = newtime;
	(void) write(1, obuf, cc);
	(void) fwrite(obuf, 1, cc, fscript);
    }
    done();
}

void doshell()
{
    char *shname;

#if 0
    int t;

    t = open(_PATH_TTY, O_RDWR);
    if (t >= 0) {
	(void) ioctl(t, TIOCNOTTY, (char *) 0);
	(void) close(t);
    }
#endif

    getslave();
    (void) close(master);
    (void) fclose(fscript);
    (void) dup2(slave, 0);
    (void) dup2(slave, 1);
    (void) dup2(slave, 2);
    (void) close(slave);

    shname = strrchr(shell, '/');
    if (shname)
	shname++;
    else
	shname = shell;

    execl(shell, shname, "-i", 0);

    perror(shell);
    fail();
}

void fixtty()
{
    struct termios rtt;

    rtt = tt;
    cfmakeraw(&rtt);
    rtt.c_lflag &= ~ECHO;
    (void) tcsetattr(0, TCSAFLUSH, &rtt);
}

void fail()
{
    (void) kill(0, SIGTERM);
    done();
}

void done()
{
    if (subchild) {
	(void) fclose(fscript);
	(void) close(master);
    } else {
	(void) tcsetattr(0, TCSAFLUSH, &tt);
    }
    exit(0);
}

void getmaster()
{
    char *pty, *bank, *cp;
    struct stat stb;

    pty = &line[strlen("/dev/ptyp")];
    for (bank = "pqrs"; *bank; bank++) {
	line[strlen("/dev/pty")] = *bank;
	*pty = '0';
	if (stat(line, &stb) < 0)
	    break;
	for (cp = "0123456789abcdef"; *cp; cp++) {
	    *pty = *cp;
	    master = open(line, O_RDWR);
	    if (master >= 0) {
		char *tp = &line[strlen("/dev/")];
		int ok;

		/* verify slave side is usable */
		*tp = 't';
		ok = access(line, R_OK | W_OK) == 0;
		*tp = 'p';
		if (ok) {
		    (void) tcgetattr(0, &tt);
		    (void) ioctl(0, TIOCGWINSZ, (char *) &win);
		    return;
		}
		(void) close(master);
	    }
	}
    }
    fprintf(stderr, "Out of pty's\n");
    fail();
}

void getslave()
{
    line[strlen("/dev/")] = 't';
    slave = open(line, O_RDWR);
    if (slave < 0) {
	perror(line);
	fail();
    }
    (void) tcsetattr(slave, TCSAFLUSH, &tt);
    (void) ioctl(slave, TIOCSWINSZ, (char *) &win);
    (void) setsid();
    (void) ioctl(slave, TIOCSCTTY, 0);
}

void mysyslog(int pri, const char *fmt, ...)
{
    va_list ap;
    char buf[BUFSIZ + 1];

    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);

    openlog(progname, 0, LOG_LOCAL2);
    syslog(pri, "%s", buf);
    closelog();
}
